#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sched.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifdef HAVE_IMLIB
# include <Imlib.h>
#endif
#ifdef HAVE_IMLIB2
# include <X11/Xlib.h>
# include <Imlib2.h>
#endif
#include <gtk/gtk.h>

#include "../include/string.h"
#include "../include/fio.h"
#include "../include/prochandle.h"

#include "libps.h"
#include "guiutils.h"
#include "cdialog.h"
#include "progressdialog.h"

#include "imgview.h"
#include "printopts.h"
#include "print.h"
#include "config.h"

#include "images/icon_print_32x32.xpm"


#ifdef HAVE_IMLIB
/* Handle used for passing to all Imlib functions, not needed for
 * Imlib2. This is foundationally declared in main.c.
 */
extern void *imlib_handle;
#endif  /* HAVE_IMLIB */


/*
 *	Progress callback data structure, this structure is used for
 *	the progress callback client data while writing postscript
 *	files.
 */
typedef struct {

	GtkWidget *ref_toplevel;
	const gchar *filename;
	const gchar *progress_message;

} ps_progress_cb_struct;


static gbool is_running(pid_t pid);
static int ps_write_progress_cb(void *data, long cur, long max);
static gint do_imlib_save(
	imgview_image_struct *img, const gchar *filename,
        gint output_x, gint output_y,
        gint paper_width, gint paper_height,
	gint paper_widthm, gint paper_heightm	/* With dpi applied. */
);
static gint iv_print_to_file(
	GtkWidget *ref_toplevel,
        imgview_struct *iv,
        print_values_struct *pv
);

gint iv_print(
        GtkWidget *ref_toplevel,
        imgview_struct *iv,
        print_values_struct *pv
);


#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define ABSOLUTE(x)     (((x) < 0) ? ((x) * -1) : (x))


/*
 *	Returns TRUE if the process is still running.
 */
static gbool is_running(pid_t pid)
{
        struct sched_param sp;


        /* Cannot be 0 (because it means itself to sched_getparam()). */
        if(pid == 0)
            return(FALSE);

        if(sched_getparam(
            pid,
            &sp
        ) == 0)
            return(TRUE);
        else
            return(FALSE);
}


/*
 *	This function is called once or more times during calls to the
 *	postscript library functions.
 *
 *	A non-zero return means to abort the operation.
 */
static int ps_write_progress_cb(void *data, long cur, long max)
{
	ps_progress_cb_struct *cb_data = (ps_progress_cb_struct *)data;
	if(cb_data == NULL)
	    return(0);

	/* If progress dialog is not mapped, then map it with initial
	 * values.
	 */
	if(!ProgressDialogIsQuery())
	{
	    ProgressDialogSetTransientFor(cb_data->ref_toplevel);
	    ProgressDialogMap(
		"Print Status",
		cb_data->progress_message,
		(const u_int8_t **)icon_print_32x32_xpm,
		"Stop"
	    );
	}

	/* Update progress. */
	ProgressDialogUpdate(
	    NULL,		/* Do not change title. */
	    NULL,		/* Do not change label. */
	    NULL,		/* Do not change icon. */
	    NULL,		/* do not change button label. */
	    ((max > 0) ? (gdouble)cur / (gdouble)max : 0.0),
	    35,
	    TRUE
	);

	/* Stop button clicked one or more times on progress dialog?
	 * If it was then -1 needs to be returned to indicate abort
	 * writing of postscript file.
	 */
	if(ProgressDialogStopCount() > 0)
	    return(-1);
	else
	    return(0);
}

/*
 *	Procedure to save the given image to file the Imlib way
 *
 *	The output_x and output_y specify the upper left corner offset
 *	in pixels of the output image.
 *
 *	The paper_width and paper_height specify the size of the
 *	paper media in pixels.
 *
 *	The paper_widthm and paper_heightm specify the size of the
 *	actual paper media (with dpi applied).
 */
static gint do_imlib_save(
        imgview_image_struct *img, const gchar *filename,
	gint output_x, gint output_y,
	gint paper_width, gint paper_height,
	gint paper_widthm, gint paper_heightm	/* With dpi applied. */
)
{
	gint status;
	gint sbpl, sbpp, width, height;
	const guint8 *src_data;


	if((img == NULL) || (filename == NULL))
	    return(-1);

	src_data = (const guint8 *)img->mem;
	if(src_data == NULL)
	    return(-1);

	sbpl = img->bpl;
	sbpp = img->bpp;
	width = img->width;
	height = img->height;

	if((width < 1) || (height < 1) || (sbpp < 1) || (sbpl < 1))
	    return(-1);

	status = -1;
#ifdef HAVE_IMLIB
	if(1)
	{
	    gint tbpl, tbpp;
	    guint8 *tar_data_rgb, *tar_data_alpha;
            ImlibImage *imlib_image;
            ImlibSaveInfo *si;


	    /* Allocate target rgb and alpha buffers. */
	    tbpp = 3;
	    tbpl = width * tbpp;
	    tar_data_rgb = (guint8 *)g_malloc(tbpl * height);
	    tar_data_alpha = (guint8 *)g_malloc(width * height);
	    if((tar_data_rgb != NULL) && (tar_data_alpha != NULL))
	    {
		gint x, y, bc, min_rgb_bpp;
		guint8 *tar_ptr;
		const guint8 *src_ptr;


		/* Calculate smaller of rgb bytes per pixel. */
		min_rgb_bpp = MIN(sbpp, tbpp);

                /* Copy source image to our target buffers. */
                for(y = 0; y < height; y++)
                {
                    for(x = 0; x < width; x++)
                    {
                        src_ptr = &src_data[
                            (y * sbpl) + (x * sbpp)
                        ];

                        /* Copy rgb. */
                        tar_ptr = &tar_data_rgb[
                            (y * tbpl) + (x * tbpp)
                        ];
                        for(bc = 0; bc < min_rgb_bpp; bc++)
                            *tar_ptr++ = *src_ptr++;

                        /* Copy alpha. */
                        tar_ptr = &tar_data_alpha[
                            (y * width) + (x * 1)
                        ];
                        /* More bytes left on source pixel pointer? */
                        if(bc < sbpp)
                        {
                            *tar_ptr = *src_ptr++;
                            bc++;
                        }
                        else
                        {
                            *tar_ptr = 0xff;
                        }
                    }
                }
            }

            /* Create Imlib image from loaded data. */
	    imlib_image = Imlib_create_image_from_data(
		imlib_handle,
		(unsigned char *)tar_data_rgb,
		(unsigned char *)tar_data_alpha,
		width, height
	    );
            if(imlib_image != NULL)
            {
                /* Need to realize changes. */
                Imlib_changed_image(imlib_handle, imlib_image);
            }

            /* Deallocate our target buffers, we don't need them
             * anymore after creating the Imlib image.
             */
            g_free(tar_data_rgb);
	    tar_data_rgb = NULL;
            g_free(tar_data_alpha);
	    tar_data_alpha = NULL;

            /* Allocate and set up save info. */
	    si = (ImlibSaveInfo *)g_malloc0(sizeof(ImlibSaveInfo));
            if(si != NULL)
            {
		int page_size_code;


                /* Quality for lossy formats (ie jpg), valid values from
                 * 0 = 0% to 256 = 100%.
                 */
                si->quality = 1.0 * 256;

                /* Scaling for paper relative formats, valid values from
                 * 0 = 0% to 1024 = 100%.
                 */
		si->scaling = 0;
/*
Can't get this to work, fuck it!
                si->scaling = (int)(1024.0 *
		    (gfloat)(width + height) /
		    (gfloat)(paper_widthm + paper_heightm)
		);
 */

                /* Bottom-left relative paper justification, valid values
                 * from 0 to 1024.
                 */
                si->xjustification = (int)(1024.0 *
		    ((gfloat)output_x / (gfloat)paper_width)
		);
                si->yjustification = (int)(1024.0 *
		    (1.0 - ((gfloat)output_y / (gfloat)paper_height))
		);

		/* Calculate paper size code from actual paper size. */
		if((paper_width == 750) && (paper_height == 1000))
		    page_size_code = PAGE_SIZE_EXECUTIVE;
		else if((paper_width == 850) && (paper_height == 1100))
		    page_size_code = PAGE_SIZE_LETTER;
		else if((paper_width == 850) && (paper_height == 1400))
                    page_size_code = PAGE_SIZE_LEGAL;
                else if((paper_width == 826) && (paper_height == 1169))
                    page_size_code = PAGE_SIZE_A4;
                else if((paper_width == 1169) && (paper_height == 1653))
                    page_size_code = PAGE_SIZE_A3;
                else if((paper_width == 583) && (paper_height == 826))
                    page_size_code = PAGE_SIZE_A5;
		else
		    page_size_code = PAGE_SIZE_LETTER;	/* All else assume letter. */
                si->page_size = page_size_code;

                /* Save as greyscale (value of 0) or color (value of 1). */
                si->color = 1;
            }

            /* Set Imlib fallback 1, meaning use ImageMagick & NETPBM. */
            Imlib_set_fallback(imlib_handle, 1);

            /* Copy new_path and save image to target path. */
            if(imlib_image != NULL)
            {
                gchar *dpath = g_strdup(filename);
                status = Imlib_save_image(
                    imlib_handle, imlib_image, dpath, si
                );
                g_free(dpath); dpath = NULL;

                /* Check if save was successful and convert status. */
                if(status == 1)
		    status = 0;
                else
		    status = -1;
	    }

            /* Destroy Imlib image, it is no longer needed in the cache. */
            if(imlib_image != NULL)
            {
                Imlib_kill_image(imlib_handle, imlib_image);
                imlib_image = NULL;
            }

            /* Deallocate save info. */
            g_free(si);
            si = NULL;
	}
#endif	/* HAVE_IMLIB */
#ifdef HAVE_IMLIB2

/* Need to work on this. */




#endif

	return(status);
}


/*
 *	Print to file procedure.
 */
static gint iv_print_to_file(
        GtkWidget *ref_toplevel,
        imgview_struct *iv,
        print_values_struct *pv
)
{
	gint status;
	gbool need_imlib_save = TRUE;
	guint vflags;
	gint output_x = 0, output_y = 0;
	gint paper_width = 320, paper_height = 240;
	gint paper_widthm = 320 * 0.72, paper_heightm = 240 * 0.72;
	const gchar *filename = NULL, *ext_ptr;
	imgview_image_struct *src_img, *tar_img = NULL;
	struct stat stat_buf;


/* Macro to deallocate all resources allocated by this function. */
#define DO_FREE_LOCALS  	\
{ \
 if(tar_img != NULL) \
 { \
  gdk_flush(); \
  ImgViewImageDelete(tar_img); \
  tar_img = NULL; \
 } \
}

	/* Get print values. */
	vflags = pv->flags;

	if(vflags & PRINT_VALUE_FILE_NAME)
	    filename = (const gchar *)pv->file_name;

	if(vflags & PRINT_VALUE_OUTPUT_X)
	    output_x = pv->output_x;
        if(vflags & PRINT_VALUE_OUTPUT_Y)
            output_y = pv->output_y;

        if((vflags & PRINT_VALUE_PAPER_WIDTH) &&
	   (vflags & PRINT_VALUE_DPI_X)
	)
	{
	    paper_width = pv->paper_width;
            paper_widthm = (gint)(
		(gfloat)paper_width * (gfloat)pv->dpi_x / 100.0
	    );
	}
        if((vflags & PRINT_VALUE_PAPER_HEIGHT) &&
           (vflags & PRINT_VALUE_DPI_Y)
        )
	{
	    paper_height = pv->paper_height;
            paper_heightm = (gint)(
                (gfloat)paper_height * (gfloat)pv->dpi_y / 100.0
            );
	}

	/* Paper size valid? */
	if((paper_width < 1) || (paper_height < 1))
	{
	    DO_FREE_LOCALS
	    return(-2);
	}

	/* Did we get a file name to save to from the print values? */
	if(filename == NULL)
	{
	    CDialogSetTransientFor(ref_toplevel);
	    CDialogGetResponse(
		"Print Failed",
		"Internal error: File name is NULL.",
		NULL,
		CDIALOG_ICON_ERROR,
		CDIALOG_BTNFLAG_OK,
		CDIALOG_BTNFLAG_OK
	    );
	    CDialogSetTransientFor(NULL);
	    DO_FREE_LOCALS
	    return(-1);
	}

	/* The extension of the file name must be one of the accepted
	 * post script formats.
	 */
	ext_ptr = strrchr(filename, '.');
	if(ext_ptr == NULL)
	    ext_ptr = "";
	/* Check extension to see if it matches atleast one of the known
	 * postscript file format extensions.
	 */
	if(strcasecmp(ext_ptr, FTYPE_EXT_PS) &&
           strcasecmp(ext_ptr, FTYPE_EXT_PS2) &&
           strcasecmp(ext_ptr, FTYPE_EXT_PS3)
	)
	{
	    /* Given file name does not have the proper extension to
	     * indicate that it is a postscript file.
	     */
            CDialogSetTransientFor(ref_toplevel);
            CDialogGetResponse(
"Print Failed",
"Invalid file type extension.",
"The file name's file type extension must indicate\n\
that it is a postscript file. Try specifying .ps as\n\
the file name's extension.",
                CDIALOG_ICON_ERROR,
                CDIALOG_BTNFLAG_OK | CDIALOG_BTNFLAG_HELP,
                CDIALOG_BTNFLAG_OK
            );
            CDialogSetTransientFor(NULL);
	    DO_FREE_LOCALS
	    return(-2);
	}

	/* Check if image already exists and if so then prompt for
	 * overwrite.
	 */
	if(!stat(filename, &stat_buf))
	{
            /* Confirm overwrite. */
            gint len = strlen(filename) + (80 * 2);
            gchar *buf;

            buf = (gchar *)g_malloc(len);
            if(buf != NULL)
                sprintf(
                    buf,
                    "Overwrite existing file:\n    %s",
                    filename
                );
	    CDialogSetTransientFor(ref_toplevel);
            status = CDialogGetResponse(
                "Confirm Overwrite",
                buf,
                NULL,
                CDIALOG_ICON_WARNING,
                CDIALOG_BTNFLAG_YES | CDIALOG_BTNFLAG_NO,
                CDIALOG_BTNFLAG_NO
            );
            CDialogSetTransientFor(NULL);

            g_free(buf);

            switch(status)
            {
              case CDIALOG_RESPONSE_YES:
              case CDIALOG_RESPONSE_YES_TO_ALL:
              case CDIALOG_RESPONSE_OK:
                break;

              default:
		DO_FREE_LOCALS
                return(-4);
                break;
            }
	}

	/* Get source image. */
	src_img = ImgViewGetImage(iv);
	if(src_img == NULL)
	{
            CDialogSetTransientFor(ref_toplevel);
            CDialogGetResponse(
                "Print Failed",
                "No image data loaded.",
                NULL,
                CDIALOG_ICON_WARNING,
                CDIALOG_BTNFLAG_OK,
                CDIALOG_BTNFLAG_OK
            );
            CDialogSetTransientFor(NULL);
	    DO_FREE_LOCALS
            return(-2);
        }


	/* Create target image for printing based on given print values.
	 * This new target image will be properly rotated and colored
	 * with respect to the given print values.
	 */
	tar_img = PrintWinCreateImageFromValues(src_img, pv);
	if(tar_img == NULL)
	{
            CDialogSetTransientFor(ref_toplevel);
            CDialogGetResponse(
                "Print Failed",
                "Could not generate target image for printing.",
                NULL,
                CDIALOG_ICON_ERROR,
                CDIALOG_BTNFLAG_OK,
                CDIALOG_BTNFLAG_OK
            );
            CDialogSetTransientFor(NULL);
	    DO_FREE_LOCALS
            return(-3);
	}


	/* Mark status as -1 to indicate error for now, it will be set
	 * to success or left as error depending on if a save was made
	 * below.
	 */
	status = -1;

	/* Write PS file. */
	while(1)
	{
	    FILE *fp;
	    const gchar *mesg;
	    PSFormatStruct fmt;
	    gchar *progress_message = NULL;
	    ps_progress_cb_struct *cb_data;


	    /* Open PS file for writing. */
	    fp = FOpen(filename, "wb");
	    if(fp == NULL)
	    {
                CDialogSetTransientFor(ref_toplevel);
                CDialogGetResponse(
                    "Print Failed",
                    "Could not open postscript file for writing.",
                    NULL,
                    CDIALOG_ICON_ERROR,
                    CDIALOG_BTNFLAG_OK,
                    CDIALOG_BTNFLAG_OK
                );
                CDialogSetTransientFor(NULL);
		status = -1;
		break;
	    }

	    /* Set up PS format strcture. */
	    fmt.flags = (PSFlagDPIX | PSFlagDPIY |
		PSFlagPaperWidth | PSFlagPaperHeight |
		PSFlagOutputX | PSFlagOutputY
	    );
	    /* DPI resolution. */
	    fmt.dpi_x = ((vflags & PRINT_VALUE_DPI_X) ?
		pv->dpi_x : 72
	    );
	    fmt.dpi_y = ((vflags & PRINT_VALUE_DPI_Y) ?
                pv->dpi_y : 72
            );
	    /* Paper size without dpi applied. */
            fmt.paper_width = paper_width;
            fmt.paper_height = paper_height;
	    /* Output image offset. */
            fmt.output_x = output_x;
            fmt.output_y = output_y;

	    /* Set up ps progress callback structure. */
	    cb_data = (ps_progress_cb_struct *)g_malloc0(sizeof(ps_progress_cb_struct));
	    if(cb_data != NULL)
	    {
		progress_message = (gchar *)g_malloc(
		    (2 * 80) + strlen(filename)
		);
		if(progress_message != NULL)
		    sprintf(
			progress_message,
"Generating postscript spool file:\n\
    %s",
			filename
		    );

		cb_data->ref_toplevel = ref_toplevel;
		cb_data->filename = filename;
		cb_data->progress_message = (const gchar *)progress_message;
	    }

	    /* Write PS file depending on source image BPP. */
	    switch(tar_img->bpp)
	    {
	      case 4:
		status = PSWriteFilePS3RGBA(
		    fp, &fmt,
		    tar_img->width, tar_img->height,
		    tar_img->bpl,
		    (const void *)tar_img->mem,
                    cb_data,
                    ps_write_progress_cb
		);
		break;

	      case 3:
		status = PSWriteFilePS3RGB(
                    fp, &fmt,
                    tar_img->width, tar_img->height,
                    tar_img->bpl,
                    (const void *)tar_img->mem,
                    cb_data,
                    ps_write_progress_cb
                );
                break;

              case 1:
                status = PSWriteFilePS3Grey(
                    fp, &fmt,
                    tar_img->width, tar_img->height,
                    tar_img->bpl,
                    (const void *)tar_img->mem,
		    cb_data,
		    ps_write_progress_cb
                );
                break;

	      default:
		status = PSError;
		break;
	    }

	    /* Close PS file. */
	    FClose(fp);
	    fp = NULL;

            /* Unmap progress dialog which might have been mapped when
	     * ps_write_progress_cb() was called.
	     */
            ProgressDialogBreakQuery(TRUE);
            ProgressDialogSetTransientFor(NULL);

	    /* Deallocate PS callback data structure. */
	    g_free(cb_data);
	    cb_data = NULL;

	    /* Deallocate progress message since it won't be accessed
	     * any more.
	     */
	    g_free(progress_message);
	    progress_message = NULL;

	    /* Check for error. */
	    mesg = NULL;
	    switch(status)
	    {
	      case PSSuccess:
		status = 0;
		break;
	      case PSError:
		mesg = "Error writing postscript file.";
		status = -1;
		break;
	      case PSBadValue:
                mesg = "Bad value used to write postscript file.";
                status = -2;
                break;
	      case PSErrorSystem:
		mesg = "Systems error occured while writing postscript file.";
		status = -3;
		break;
	      case PSAbort:
		status = -4;
		break;
	      case PSErrorInternal:
		mesg = "Systems error occured while writing postscript file.";
		status = -3;		/* Treat as systems error. */
		break;
	      default:
		status = -1;
		break;
	    }

	    if(mesg != NULL)
	    {
                CDialogSetTransientFor(ref_toplevel);
                CDialogGetResponse(
                    "Print Failed",
                    mesg,
                    NULL,
                    CDIALOG_ICON_ERROR,
                    CDIALOG_BTNFLAG_OK,
                    CDIALOG_BTNFLAG_OK
                );
                CDialogSetTransientFor(NULL);
	    }

	    /* Break when we reach the end of this big while() loop. */
	    break;
	}
	need_imlib_save = FALSE;	/* Don't ever save using Imlib. */


        /* Need to use Imlib to save the image? (we really don't need
	 * this one anymore)
	 */
        if(need_imlib_save)
	    status = do_imlib_save(
		tar_img, filename,
		output_x, output_y,
		paper_width, paper_height,
		paper_widthm, paper_heightm
	    );

        /* Deallocate all resources allocated by this function. */
	DO_FREE_LOCALS
#undef DO_FREE_LOCALS

	return(status);
}


/*
 *	Print to printer procedure.
 */
static gint iv_print_to_printer(
        GtkWidget *ref_toplevel,
        imgview_struct *iv,
        print_values_struct *pv
)
{
        gint status;
        guint vflags;
        gint output_x = 0, output_y = 0;
        gint paper_width = 320, paper_height = 240;
        gint paper_widthm = 320 * 0.72, paper_heightm = 240 * 0.72;
        const gchar *orig_command = NULL;
	gchar *filename = NULL, *command = NULL;
        imgview_image_struct *src_img, *tar_img = NULL;
	pid_t pid;
	struct stat stat_buf;


/* Macro to deallocate all resources allocated by this function. */
#define DO_FREE_LOCALS	\
{ \
 /* Delete output image if exists. */ \
 if(tar_img != NULL) \
 { \
  gdk_flush(); \
  ImgViewImageDelete(tar_img); \
  tar_img = NULL; \
 } \
\
 /* Remove tempory postscript file if exists. */ \
 if(filename != NULL) \
 { \
  if(!stat(filename, &stat_buf)) \
  { \
   unlink(filename); \
  } \
 } \
\
 g_free(command); \
 command = NULL; \
 g_free(filename); \
 filename = NULL; \
}

        /* Get print values. */
        vflags = pv->flags;

        if(vflags & PRINT_VALUE_PRINT_COMMAND)
            orig_command = (const gchar *)pv->print_command;

        if(vflags & PRINT_VALUE_OUTPUT_X)
            output_x = pv->output_x;
        if(vflags & PRINT_VALUE_OUTPUT_Y)
            output_y = pv->output_y;

        if((vflags & PRINT_VALUE_PAPER_WIDTH) &&
           (vflags & PRINT_VALUE_DPI_X)
        )
        {
            paper_width = pv->paper_width;
            paper_widthm = (gint)(
                (gfloat)paper_width * (gfloat)pv->dpi_x / 100.0
            );
        }
        if((vflags & PRINT_VALUE_PAPER_HEIGHT) &&
           (vflags & PRINT_VALUE_DPI_Y)
        )
        {
            paper_height = pv->paper_height;
            paper_heightm = (gint)(
                (gfloat)paper_height * (gfloat)pv->dpi_y / 100.0
            );
        }

        /* Paper size valid? */
        if((paper_width < 1) || (paper_height < 1))
	{
	    DO_FREE_LOCALS
            return(-2);
	}

        /* Get source image. */
        src_img = ImgViewGetImage(iv);
        if(src_img == NULL)
        {
            CDialogSetTransientFor(ref_toplevel);
            CDialogGetResponse(
                "Print Failed",
                "No image data loaded.",
                NULL,
                CDIALOG_ICON_WARNING,
                CDIALOG_BTNFLAG_OK,
                CDIALOG_BTNFLAG_OK
            );
            CDialogSetTransientFor(NULL);
	    DO_FREE_LOCALS
            return(-2);
        }

        /* Did we get a command to use as the print command? */
        if(orig_command == NULL)
        {
            CDialogSetTransientFor(ref_toplevel);
            CDialogGetResponse(
                "Print Failed",
                "Internal error: Print command is NULL.",
                NULL,
                CDIALOG_ICON_ERROR,
                CDIALOG_BTNFLAG_OK,
                CDIALOG_BTNFLAG_OK
            );
            CDialogSetTransientFor(NULL);
	    DO_FREE_LOCALS
            return(-1);
        }

	/* Begin formulating a tempory file name and the print command. */

	/* Generate tempory file name. */
	if(filename == NULL)
	{
	    /* Generate unique file name for postscript file. */
	    filename = (gchar *)tempnam(
		NULL,		/* Let system figure out temp dir. */
		"print"
	    );
	}
	/* Failed to generate file name? */
	if(filename == NULL)
	{
            CDialogSetTransientFor(ref_toplevel);
            CDialogGetResponse(
                "Print Failed",
                "Internal error: Cannot generate tempory file name.",
                NULL,
                CDIALOG_ICON_ERROR,
                CDIALOG_BTNFLAG_OK,
                CDIALOG_BTNFLAG_OK
            );
	    CDialogSetTransientFor(NULL);
            DO_FREE_LOCALS
            return(-1);
	}
	/* Check if file name already exists? */
	if(!stat(filename, &stat_buf))
	{
	    /* It already exists, this should not happen. Deallocate 
	     * filename and return. Note that filename needs to be
	     * deallocated before calling DO_FREE_LOCALS, otherwise the
	     * actual file would get removed.
	     */
	    g_free(filename);
	    filename = NULL;
	    DO_FREE_LOCALS
	    return(-3);
	}
	/* Generate print command. */
	if(command == NULL)
	{
	    gint len = strlen(orig_command) + strlen(filename) + 80;

	    command = (gchar *)g_malloc(len);
	    if(command != NULL)
	    {
		(*command) = '\0';
		strcat(command, orig_command);
                strcat(command, " ");
                strcat(command, filename);
	    }
	}
        /* Failed to generate print command? */
        if(command == NULL)
        {
            CDialogSetTransientFor(ref_toplevel);
            CDialogGetResponse(
                "Print Failed",
                "Internal error: Cannot generate print command.",
                NULL,
                CDIALOG_ICON_ERROR,
                CDIALOG_BTNFLAG_OK,
                CDIALOG_BTNFLAG_OK
            );
            CDialogSetTransientFor(NULL);
            DO_FREE_LOCALS
            return(-1);
        }


        /* Create target image for printing based on given print values.
         * This new target image will be properly rotated and colored
         * with respect to the given print values.
         */
        tar_img = PrintWinCreateImageFromValues(src_img, pv);
        if(tar_img == NULL)
        {
            CDialogSetTransientFor(ref_toplevel);
            CDialogGetResponse(
                "Print Failed",
                "Could not generate target image for printing.",
                NULL,
                CDIALOG_ICON_ERROR,
                CDIALOG_BTNFLAG_OK,
                CDIALOG_BTNFLAG_OK
            );
            CDialogSetTransientFor(NULL);
            DO_FREE_LOCALS
            return(-3);
        }


        /* Mark status as -1 to indicate error for now, it will be set
         * to success or left as error depending on if a save was made
         * below.
         */
	status = -1;

        /* Write PS file. */
        while(1)
        {
            FILE *fp;
            const gchar *mesg;
            PSFormatStruct fmt;
	    gchar *progress_message = NULL;
            ps_progress_cb_struct *cb_data;


            /* Open PS file for writing. */
            fp = FOpen(filename, "wb");
            if(fp == NULL)
            {
                CDialogSetTransientFor(ref_toplevel);
                CDialogGetResponse(
                    "Print Failed",
                    "Could not open postscript file for writing.",
                    NULL,
                    CDIALOG_ICON_ERROR,
                    CDIALOG_BTNFLAG_OK,
                    CDIALOG_BTNFLAG_OK
                );
                CDialogSetTransientFor(NULL);
                status = -1;
                break;
            }

            /* Set up PS format strcture. */
            fmt.flags = (PSFlagDPIX | PSFlagDPIY |
                PSFlagPaperWidth | PSFlagPaperHeight |
                PSFlagOutputX | PSFlagOutputY
            );
            /* DPI resolution. */
            fmt.dpi_x = ((vflags & PRINT_VALUE_DPI_X) ?
                pv->dpi_x : 72
            );
            fmt.dpi_y = ((vflags & PRINT_VALUE_DPI_Y) ?
                pv->dpi_y : 72
            );
            /* Paper size without dpi applied. */
            fmt.paper_width = paper_width;
            fmt.paper_height = paper_height;
            /* Output image offset. */
            fmt.output_x = output_x;
            fmt.output_y = output_y;

            /* Set up ps progress callback structure. */
            cb_data = (ps_progress_cb_struct *)g_malloc0(sizeof(ps_progress_cb_struct));
            if(cb_data != NULL)
            {
                progress_message = (gchar *)g_malloc(
                    (2 * 80) + strlen(filename)
                );
                if(progress_message != NULL)
                    sprintf(
                        progress_message,
"Generating postscript spool file:\n\
    %s",
                        filename
                    );

                cb_data->ref_toplevel = ref_toplevel;
                cb_data->filename = filename;
                cb_data->progress_message = (const gchar *)progress_message;
            }

            /* Write PS file depending on source image BPP. */
            switch(tar_img->bpp)
            {
              case 4:
                status = PSWriteFilePS3RGBA(
                    fp, &fmt,
                    tar_img->width, tar_img->height,
                    tar_img->bpl,
                    (const void *)tar_img->mem,
                    cb_data,
                    ps_write_progress_cb
                );
                break;

              case 3:
                status = PSWriteFilePS3RGB(
                    fp, &fmt,
                    tar_img->width, tar_img->height,
                    tar_img->bpl,
                    (const void *)tar_img->mem,
                    cb_data,
                    ps_write_progress_cb
                );
                break;

              case 1:
                status = PSWriteFilePS3Grey(
                    fp, &fmt,
                    tar_img->width, tar_img->height,
                    tar_img->bpl,
                    (const void *)tar_img->mem,
                    cb_data,
                    ps_write_progress_cb
                );
                break;

              default:
                status = PSError;
                break;
            }

            /* Close PS file. */
            FClose(fp);
            fp = NULL;

            /* Unmap progress dialog which might have been mapped when
             * ps_write_progress_cb() was called.
             */
            ProgressDialogBreakQuery(TRUE);
            ProgressDialogSetTransientFor(NULL);

            /* Deallocate PS callback data structure. */
            g_free(cb_data);
            cb_data = NULL;

            /* Deallocate progress message since it won't be accessed
             * any more.
             */
            g_free(progress_message);
            progress_message = NULL;

            /* Check for error. */
            mesg = NULL;
            switch(status)
            {
              case PSSuccess:
                status = 0;
                break;
              case PSError:
                mesg = "Error writing postscript file.";
                status = -1;
                break;
              case PSBadValue:
                mesg = "Bad value used to write postscript file.";
                status = -2;
                break;
              case PSErrorSystem:
                mesg = "Systems error occured while writing postscript file.";
                status = -3;
                break;
              case PSAbort:
                status = -4;
                break;
              case PSErrorInternal:
                mesg = "Systems error occured while writing postscript file.";
                status = -3;            /* Treat as systems error. */
                break;
              default:
                status = -1;
                break;
            }
            if(mesg != NULL)
            {
                CDialogSetTransientFor(ref_toplevel);
                CDialogGetResponse(
                    "Print Failed",
                    mesg,
                    NULL,
                    CDIALOG_ICON_ERROR,
                    CDIALOG_BTNFLAG_OK,
                    CDIALOG_BTNFLAG_OK
                );
                CDialogSetTransientFor(NULL);
            }

            /* Break when we reach the end of this big while() loop. */
            break;
        }

	/* Deallocate target image, it is not needed anymore. */
	if(tar_img != NULL)
	{
	    gdk_flush();
	    ImgViewImageDelete(tar_img);
	    tar_img = NULL;
	}

	/* Failed to write PS image? */
	if(status == -4)
	{
	    /* Aborted, which means print procedure must be aborted. */
            CDialogSetTransientFor(ref_toplevel);
            CDialogGetResponse(
		"Print Aborted",
"User aborted postscript file generation, print procedure\n\
has also been aborted.",
                NULL,
                CDIALOG_ICON_WARNING,
                CDIALOG_BTNFLAG_OK,
                CDIALOG_BTNFLAG_OK
            );
            CDialogSetTransientFor(NULL);
	    DO_FREE_LOCALS
	    return(0);
	}
	if(status)
	{
	    /* Other error occured while writing postscript image. */
            DO_FREE_LOCALS
            return(-1);
	}


	/* Run print command. */
	pid = Exec(command);
	if(pid <= 0)
	{
            CDialogSetTransientFor(ref_toplevel);
            CDialogGetResponse(
                "Print Failed",
                "Error occured while executing print command.",
                NULL,
                CDIALOG_ICON_ERROR,
                CDIALOG_BTNFLAG_OK,
                CDIALOG_BTNFLAG_OK
            );
            CDialogSetTransientFor(NULL);
	    DO_FREE_LOCALS
	    return(-1);
	}


	/* Map progress dialog and wait for print process to finish. */

	/* Set up and map progress dialog. */
	ProgressDialogSetTransientFor(ref_toplevel);
	/* While print process is running... */
	while(is_running(pid))
	{
	    /* If progress dialog is not mapped, then map it with initial
	     * values.
	     */
	    if(!ProgressDialogIsQuery())
		ProgressDialogMap(
		    "Print Status",
		    "Printing...",
                    (const u_int8_t **)icon_print_32x32_xpm,
                    "Stop"
		);

	    /* Update progress. */
            ProgressDialogUpdateUnknown(
                "Print Status",
                "Printing...",
                (const u_int8_t **)icon_print_32x32_xpm,
                "Stop",
                TRUE
            );

            /* User abort? */
            if(ProgressDialogStopCount() > 0)
                break;
	}
	/* Unmap progress dialog. */
	ProgressDialogBreakQuery(TRUE);
	ProgressDialogSetTransientFor(NULL);

	/* Process still running? kill it. */
	if(is_running(pid))
	    kill(pid, SIGINT);
	pid = 0;

        /* Deallocate all resources allocated by this function. */
	DO_FREE_LOCALS
#undef DO_FREE_LOCALS

        return(0);
}

/*
 *	Print procedure.
 *
 *	Return values:
 *
 *	0	Success
 *	-1	General error
 *	-2	Bad value or missing value
 *	-3	Systems error
 *	-4	User aborted
 */
gint iv_print(
        GtkWidget *ref_toplevel,
        imgview_struct *iv,
        print_values_struct *pv
)
{
	gint status;
	guint vflags;
	gint output_type = PRINT_OUTPUT_TO_PRINTER;


	if((iv == NULL) || (pv == NULL))
	    return(-1);

#ifdef HAVE_IMLIB
	if(imlib_handle == NULL)
	    return(-3);
#endif

	/* Check if we want to print to printer or file. */
	vflags = pv->flags;
	if(vflags & PRINT_VALUE_OUTPUT_TYPE)
	    output_type = pv->output_type;

	if(output_type == PRINT_OUTPUT_TO_FILE)
	{
	    /* Call print to file procedure. */
	    status = iv_print_to_file(ref_toplevel, iv, pv);
	}
	else
	{
	    /* Call print to printer procedure. */
            status = iv_print_to_printer(ref_toplevel, iv, pv);
	}

	return(status);
}
